function scrollIntoView(element: HTMLDivElement, spot: {
  left?: number,
  top?: number
}, scrollMatches = false) {
  // Assuming offsetParent is available (it's not available when viewer is in
  // hidden iframe or object). We have to scroll: if the offsetParent is not set
  // producing the error. See also animationStarted.
  let parent = element.offsetParent as HTMLDivElement
  if (!parent) {
    console.error("offsetParent is not set -- cannot scroll")
    return
  }
  let offsetY = element.offsetTop + element.clientTop
  let offsetX = element.offsetLeft + element.clientLeft
  while (
    (parent.clientHeight === parent.scrollHeight &&
      parent.clientWidth === parent.scrollWidth) ||
    (scrollMatches && getComputedStyle(parent).overflow === "hidden")
  ) {
    offsetY += parent.offsetTop
    offsetX += parent.offsetLeft

    parent = parent.offsetParent as HTMLDivElement
    if (!parent) {
      return // no need to scroll
    }
  }
  if (spot) {
    if (spot.top !== undefined) {
      offsetY += spot.top
    }
    if (spot.left !== undefined) {
      offsetX += spot.left
      parent.scrollLeft = offsetX
    }
  }
  parent.scrollTop = offsetY
}

class PromiseCapability<T> {
  #settled: boolean = false
  promise: Promise<T>
  resolve!: (data: T | PromiseLike<T>) => void
  reject!: (reason: any) => void

  constructor() {
    this.promise = new Promise((resolve, reject) => {
      this.resolve = (data: T | PromiseLike<T>) => {
        this.#settled = true
        resolve(data)
      }

      this.reject = (reason: any) => {
        this.#settled = true
        reject(reason)
      }
    })
  }

  get settled(): boolean {
    return this.#settled
  }
}

function getPdfFilenameFromUrl(url: string | unknown, defaultFilename = "document.pdf") {
  if (typeof url !== "string") {
    return defaultFilename
  }

  const uriMatch = /^(?:(?:[^:]+:)?\/\/[^/]+)?([^?#]*)(\?[^#]*)?(#.*)?$/.exec(url)
  if (!uriMatch) {
    return defaultFilename
  }

  const [, path, query, ref] = uriMatch

  const filenameMatch = /[^/?#=]+\.pdf\b(?!.*\.pdf\b)/i.exec(path + query + ref)

  if (!filenameMatch) {
    return defaultFilename
  }

  let suggestedFilename = filenameMatch[0]

  if (suggestedFilename.includes("%")) {
    try {
      suggestedFilename = /[^/?#=]+\.pdf\b(?!.*\.pdf\b)/i.exec(decodeURIComponent(suggestedFilename))![0]
    } catch {
      // Ignore errors in case of malformed URIs
    }
  }

  return suggestedFilename || defaultFilename
}

export {
  scrollIntoView,
  PromiseCapability,
  getPdfFilenameFromUrl,
}
